#include <Wire.h>
#include "MAX30105.h"
#include <ArduinoBLE.h>
#include "heartRate.h"
#include <LSM6DS3.h>

BLEService deviceService("19B10000-E8F2-537E-4F6C-D104768A1215");
BLEStringCharacteristic heartChar("19B10001-E8F2-537E-4F6C-D104768A1215", BLERead | BLENotify, 10);
BLEStringCharacteristic stepChar("19B10002-E8F2-537E-4F6C-D104768A1215", BLERead | BLENotify, 10);
BLEStringCharacteristic tempChar("19B10003-E8F2-537E-4F6C-D104768A1215", BLERead | BLENotify, 10);

MAX30105 particleSensor;
LSM6DS3 myIMU(I2C_MODE, 0x6A);

const byte RATE_SIZE = 4;
byte rates[RATE_SIZE];
byte rateSpot = 0;
long lastBeat = 0;

const int numReadings = 8;
const float accelerationThreshold = 3.8;
const unsigned long stepCheckInterval = 400;
unsigned long previousTime = 0;
int stepCount = 0;
int stepsSinceLastTransmission = 0;

float beatsPerMinute;
int beatAvg;
float readingsX[numReadings];
float readingsY[numReadings];
float readingsZ[numReadings];
int idx = 0;
float totalX = 0;
float totalY = 0;
float totalZ = 0;

void setup() {
  Serial.begin(9600);
  Serial.println("Initializing...");
  while (!Serial);

  if (!BLE.begin()) {
    Serial.println("starting BLE failed!");
    while (1);
  }
  while (!Serial);

  if (myIMU.begin() != 0) {
    Serial.println("Device error");
    while (1);
  } else {
    Serial.println("aX,aY,aZ,gX,gY,gZ");
  }

  BLE.setLocalName("health");
  BLE.setAdvertisedService(deviceService);
  deviceService.addCharacteristic(heartChar);
  deviceService.addCharacteristic(stepChar);
  deviceService.addCharacteristic(tempChar);
  BLE.addService(deviceService);
  heartChar.writeValue("0");
  stepChar.writeValue("0");
  tempChar.writeValue("0");
  BLE.advertise();
  Serial.println("Bluetooth® device active, waiting for connections...");

  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, LOW);

  if (!particleSensor.begin(Wire, I2C_SPEED_FAST)) {
    Serial.println("MAX30105 was not found. Please check wiring/power. ");
    while (1);
  }
  Serial.println("Place your index finger on the sensor with steady pressure.");

  particleSensor.setup();
  particleSensor.setPulseAmplitudeRed(0x0A);
  particleSensor.setPulseAmplitudeGreen(0);
}

void loop() {
  BLEDevice central = BLE.central();
  if (central) {
    Serial.print("Connected to central: ");
    Serial.println(central.address());
    digitalWrite(LED_BUILTIN, HIGH);

    while (central.connected()) {
      long currentMillis = millis();
      updateHeart();
      updateStep();
      updateTemperature();
    }

    digitalWrite(LED_BUILTIN, LOW);
    Serial.print(F("Disconnected from central: "));
    Serial.println(central.address());
  }
}

void updateStep() {
  float aX = myIMU.readFloatAccelX();
  float aY = myIMU.readFloatAccelY();
  float aZ = myIMU.readFloatAccelZ();

  totalX = totalX - readingsX[idx] + aX;
  totalY = totalY - readingsY[idx] + aY;
  totalZ = totalZ - readingsZ[idx] + aZ;
  readingsX[idx] = aX;
  readingsY[idx] = aY;
  readingsZ[idx] = aZ;
  idx = (idx + 1) % numReadings;

  float averageX = totalX / numReadings;
  float averageY = totalY / numReadings;
  float averageZ = totalZ / numReadings;

  float accelerationMagnitude = sqrt(averageX * averageX + averageY * averageY + averageZ * averageZ);

  unsigned long currentTime = millis();
  if (currentTime - previousTime >= stepCheckInterval) {
    if (accelerationMagnitude > accelerationThreshold) {
      stepCount++;
      stepsSinceLastTransmission++;
      Serial.print("Step Detected! Step Count: ");
      Serial.println(stepCount);
    } else {
      Serial.println("No Step Detected");
    }
    previousTime = currentTime;

    if (stepsSinceLastTransmission >= 10) {
      stepChar.writeValue(String(stepCount));
      Serial.println("Step count sent over Bluetooth.");
      stepsSinceLastTransmission = 0;
    }
  }

  delay(10);
}

void updateHeart() {
  long irValue = particleSensor.getIR();

  if (checkForBeat(irValue) == true) {
    long delta = millis() - lastBeat;
    lastBeat = millis();

    beatsPerMinute = 60 / (delta / 1000.0);

    if (beatsPerMinute < 255 && beatsPerMinute > 20) {
      rates[rateSpot++] = (byte)beatsPerMinute;
      rateSpot %= RATE_SIZE;

      beatAvg = 0;
      for (byte x = 0; x < RATE_SIZE; x++)
        beatAvg += rates[x];
      beatAvg /= RATE_SIZE;
    }
  }

  Serial.print("IR=");
  Serial.print(irValue);
  Serial.print(", BPM=");
  Serial.print(beatsPerMinute);
  Serial.print(", Avg BPM=");
  Serial.print(beatAvg);

  heartChar.writeValue(String(beatAvg));

  if (irValue < 50000)
    Serial.print(" No finger?");

  Serial.println();
}

void updateTemperature() {
  float temperature = myIMU.readTempC();
  Serial.print("Temperature: ");
  Serial.println(temperature);
  tempChar.writeValue(String(temperature));
}
